@@ -1,7 +1,7 @@
 package Bencher::Backend;
 
 our $DATE = '2016-03-15'; # DATE
-our $VERSION = '0.38'; # VERSION
+our $VERSION = '0.39'; # VERSION
 
 use 5.010001;
 use strict;
@@ -134,6 +134,7 @@ sub _get_scenario {
     my $scenario;
     if (defined $pargs->{scenario_file}) {
         $scenario = do $pargs->{scenario_file};
+        die "Can't load scenario file '$pargs->{scenario_file}': $@" if $@;
     } elsif (defined $pargs->{scenario_module}) {
         my $m = "Bencher::Scenario::$pargs->{scenario_module}"; $m =~ s!/!::!g;
         my $mp = $m; $mp =~ s!::!/!g; $mp .= ".pm";
@@ -181,7 +182,8 @@ sub _parse_scenario {
             my $p = { %$p0, seq=>$i };
             $p->{include_by_default} //= 1;
             $p->{type} //= do {
-                if ($p->{cmdline} || $p->{cmdline_template}) {
+                if ($p->{cmdline} || $p->{cmdline_template} ||
+                        $p->{perl_cmdline} || $p->{perl_cmdline_template}) {
                     'command';
                 } else {
                     'perl_code';
@@ -344,15 +346,15 @@ sub _gen_items {
     $parsed->{items} = [];
     my @permute;
 
-    # XXX allow permutation of perl path
     # XXX allow permutation of module path
 
     my $participants;
     my $datasets;
     my $module_startup = $pargs->{module_startup} // $parsed->{module_startup};
 
+    my @modules = _get_participant_modules($parsed);
+
     if ($module_startup) {
-        my @modules = _get_participant_modules($parsed);
 
         return [412, "There are no modules to benchmark ".
                     "the startup overhead of"]
@@ -364,7 +366,7 @@ sub _gen_items {
                 seq  => 0,
                 name => "perl -e1 (baseline)",
                 type => 'command',
-                cmdline => [$^X, "-e1"],
+                perl_cmdline => ["-e1"],
             };
 
             my $i = 0;
@@ -374,7 +376,7 @@ sub _gen_items {
                     seq  => $i,
                     name => $mod,
                     type => 'command',
-                    cmdline => [$^X, "-M$mod", "-e1"],
+                    perl_cmdline => ["-M$mod", "-e1"],
                 };
             }
         }
@@ -386,6 +388,81 @@ sub _gen_items {
         $datasets = $parsed->{datasets} if $parsed->{datasets};
     }
 
+    my %perl_exes; # key=name, val=path
+    {
+        my @perls;
+        if ($pargs->{multiperl}) {
+            require App::perlbrew;
+            my $pb = App::perlbrew->new;
+            @perls = $pb->installed_perls;
+            if ($pargs->{include_perls} && @{ $pargs->{include_perls} }) {
+                @perls = grep {
+                    my $p = $_;
+                    (grep { $p->{name} eq $_ } @{ $pargs->{include_perls} }) ? 1:0;
+                } @perls;
+            }
+            if ($pargs->{exclude_perls} && @{ $pargs->{exclude_perls} }) {
+                @perls = grep {
+                    my $p = $_;
+                    (grep { $p->{name} eq $_ } @{ $pargs->{exclude_perls} }) ? 0:1;
+                } @perls;
+            }
+            die "You have to include at least one perl\n" unless @perls;
+            for (@perls) {
+                $perl_exes{$_->{name}} = $_->{executable};
+            }
+            @perls = map {$_->{name}} @perls;
+        } else {
+            $perl_exes{perl} = $^X;
+            @perls = ("perl");
+        }
+        push @permute, "perl", \@perls;
+    }
+
+    my %perl_opts; # key=name, val=[opt, ...]
+    if ($pargs->{multimodver}) {
+        require ExtUtils::MakeMaker;
+        require Module::Path::More;
+
+        return [412, "There are no modules to search multiple versions of"]
+            unless @modules;
+
+        return [412, "Module '$pargs->{multimodver}' is not among modules to benchmark"]
+            unless grep {$pargs->{multimodver} eq $_} @modules;
+
+        local @INC = @INC;
+        if ($pargs->{include_path} && @{ $pargs->{include_path} }) {
+            unshift @INC, $_ for reverse @{ $pargs->{include_path} };
+        }
+
+        my %versions; # key=module name
+        my $paths = Module::Path::More::module_path(module=>$pargs->{multimodver}, all=>1);
+
+        if (@$paths < 1) {
+            return [412, "Can't find module '$pargs->{multimodver}', try adding some --include-path"];
+        } elsif (@$paths == 1) {
+            return [412, "Only found one path for module '$pargs->{multimodver}', try adding some --include-path"];
+        }
+        for my $path (@$paths) {
+            my $v = MM->parse_version($path);
+            $v = undef if defined($v) && $v eq 'undef';
+            if (!defined($v)) {
+                $log->warnf("Can't parse version from %s", $path);
+                next;
+            }
+            $versions{$v}++;
+            my $incdir = $path;
+            my $mod_pm = $pargs->{multimodver}; $mod_pm =~ s!::!/!g; $mod_pm .= ".pm";
+            $incdir =~ s!/\Q$mod_pm\E$!!;
+            $perl_opts{$v} //= ["-I$incdir"];
+        }
+        return [412, "Can't find version number for module '$pargs->{multimodver}'"]
+            unless keys(%versions);
+        return [412, "Only found one version of module '$pargs->{multimodver}', try adding some --include-path"]
+            unless keys(%versions) > 1;
+        push @permute, "modver", [keys %versions];
+    }
+
     push @permute, "participant", [map {$_->{seq}} @$participants];
 
     if ($datasets && @$datasets) {
@@ -523,9 +600,10 @@ sub _gen_items {
 
             my $code;
             if ($p->{type} eq 'command') {
+                require String::ShellQuote;
                 my @cmd;
                 my $shell;
-                if ($p->{cmdline}) {
+                if (defined $p->{cmdline}) {
                     if (ref($p->{cmdline}) eq 'ARRAY') {
                         @cmd = @{ $p->{cmdline} };
                         $shell = 0;
@@ -533,7 +611,24 @@ sub _gen_items {
                         @cmd = ($p->{cmdline});
                         $shell = 1;
                     }
-                } else { # cmdline_template
+                } elsif (defined $p->{perl_cmdline}) {
+                    if (ref($p->{perl_cmdline}) eq 'ARRAY') {
+                        @cmd = ($perl_exes{$h->{perl}},
+                                ($h->{modver} ? @{$perl_opts{$h->{modver}}} : ()),
+                                @{ $p->{perl_cmdline} });
+                        $shell = 0;
+                    } else {
+                        @cmd = (
+                            join(
+                                " ",
+                                $perl_exes{$h->{perl}},
+                                ($h->{modver} ? map {String::ShellQuote::shell_quote($_)} @{$perl_opts{$h->{modver}}} : ()),
+                                $p->{perl_cmdline},
+                            )
+                        );
+                        $shell = 1;
+                    }
+                } elsif (defined $p->{cmdline_template}) {
                     my $template_vars;
                     if ($ds->{args}) {
                         $template_vars = { %$args };
@@ -549,24 +644,58 @@ sub _gen_items {
                         } @{ $p->{cmdline_template} };
                         $shell = 0;
                     } else {
-                        require String::ShellQuote;
                         my $cmd = $p->{cmdline_template};
                         $cmd =~ s/\<(\w+(?::\w+)?)\>/String::ShellQuote::shell_quote($template_vars->{$1})/eg;
                         @cmd = ($cmd);
                         $shell = 1;
                     }
+                } elsif (defined $p->{perl_cmdline_template}) {
+                    my $template_vars;
+                    if ($ds->{args}) {
+                        $template_vars = { %$args };
+                    } elsif ($ds->{argv}) {
+                        $template_vars = { map {$_=>$ds->{argv}[$_]}
+                                               0..$#{ $ds->{argv} } };
+                    }
+                    if (ref($p->{perl_cmdline_template}) eq 'ARRAY') {
+                        @cmd = (
+                            $perl_exes{$h->{perl}},
+                            ($h->{modver} ? @{$perl_opts{$h->{modver}}} : ()),
+                            map {
+                                my $el = $_;
+                                $el =~ s/\<(\w+(?::\w+)?)\>/$template_vars->{$1}/g;
+                                $el;
+                            } @{ $p->{perl_cmdline_template} }
+                        );
+                        $shell = 0;
+                    } else {
+                        my $cmd = join(
+                            " ",
+                            $perl_exes{$h->{perl}},
+                            ($h->{modver} ? map {String::ShellQuote::shell_quote($_)} @{$perl_opts{$h->{modver}}} : ()),
+                            $p->{perl_cmdline_template},
+                        );
+                        $cmd =~ s/\<(\w+(?::\w+)?)\>/String::ShellQuote::shell_quote($template_vars->{$1})/eg;
+                        @cmd = ($cmd);
+                        $shell = 1;
+                    }
+                } else {
+                    die "BUG: Unknown command type";
                 }
 
                 $log->debugf("Item #%d: cmdline=%s", $item_seq, \@cmd);
 
-                $code = sub {
+                {
+                    my $code_str = "sub { ";
                     if ($shell) {
-                        system $cmd[0];
+                        $code_str .= "system ".dmp($cmd[0])."; ";
                     } else {
-                        system {$cmd[0]} @cmd;
+                        $code_str .= "system {".dmp($cmd[0])."} \@{".dmp(\@cmd)."}; ";
                     }
-                    die "Command failed (child error=$?, os error=$!)\n"
-                        if $?;
+                    $code_str .= q[die "Command failed (child error=$?, os error=$!)\\n" if $?];
+                    $code_str .= "}";
+                    $code = eval $code_str;
+                    die "BUG: Can't produce code for cmdline: $@ (code string is: $code_str)" if $@;
                 };
             } elsif ($p->{type} eq 'perl_code') {
                 if ($p->{code}) {
@@ -611,7 +740,13 @@ sub _gen_items {
                 ((_permute_args => $h_args) x !!$ds->{args}),
             };
             for my $k (keys %$h) {
-                if ($k eq 'dataset') {
+                if ($k eq 'perl') {
+                    $item->{perl} = $h->{$k};
+                    $item->{_perl_exe} = $perl_exes{ $h->{$k} };
+                } elsif ($k eq 'modver') {
+                    $item->{modver} = $h->{$k};
+                    $item->{_perl_opts} = $perl_opts{ $h->{$k} };
+                } elsif ($k eq 'dataset') {
                     $item->{"dataset"} = $ds->{name} // "#$ds->{seq}";
                 } elsif ($k eq 'participant') {
                     $item->{"participant"} = $p->{name} // $p->{_name} // "#$p->{seq}";
@@ -880,6 +1015,22 @@ sub _complete_item {
     );
 }
 
+sub _complete_perl {
+    require Complete::Util;
+
+    my %args = @_;
+    my $word    = $args{word} // '';
+
+    eval { require App::perlbrew; 1 };
+    return undef if $@;
+
+    my $pb = App::perlbrew->new;
+    my @perls = $pb->installed_perls;
+    local $Complete::Common::OPT_FUZZY = 0;
+    Complete::Util::complete_array_elem(
+        word => $word, array => [map {$_->{name}} @perls]);
+}
+
 my $_alias_spec_add_dataset = {
     summary => 'Add a dataset',
     code => sub {
@@ -984,6 +1135,8 @@ $SPEC{bencher} = {
         # XXX precision & precision_limit is only relevant when action=bench
         # XXX note is only relevant when action=bench
         # XXX sort is only relevant when action=bench and format=text
+        # XXX include_perls & exclude_perls are only relevant when multiperl=1
+        # XXX include_path only relevant when multimodver=1
     },
     args => {
         scenario_file => {
@@ -994,7 +1147,7 @@ Perl file will be do()'ed and the last expression should be a hash containing
 the scenario specification.
 
 _
-            schema => 'str*',
+            schema => ['str*'],
             cmdline_aliases => {f=>{}},
         },
         scenario_module => {
@@ -1018,7 +1171,7 @@ _
         participants => {
             'summary' => 'Add participants',
             'x.name.is_plural' => 1,
-            schema => ['array*', of=>'hash*'],
+            schema => ['array*', of=>['hash*']],
             cmdline_aliases => {
                 participant => $_alias_spec_add_participant,
                 p => $_alias_spec_add_participant,
@@ -1027,7 +1180,7 @@ _
         datasets => {
             summary => 'Add datasets',
             'x.name.is_plural' => 1,
-            schema => ['array*', of=>'hash*'],
+            schema => ['array*', of=>['hash*']],
             cmdline_aliases => {
                 dataset => $_alias_spec_add_dataset,
                 d => $_alias_spec_add_dataset,
@@ -1131,7 +1284,7 @@ _
         },
         raw => {
             summary => 'Show "raw" data',
-            schema => 'bool',
+            schema => ['bool'],
             description => <<'_',
 
 When action=show-items-result, will print result as-is instead of dumping as
@@ -1145,7 +1298,7 @@ _
             tags => ['category:action'],
         },
         detail => {
-            schema => 'bool*',
+            schema => ['bool*'],
             cmdline_aliases => {l=>{}},
         },
 
@@ -1153,7 +1306,7 @@ _
             'x.name.is_plural' => 1,
             summary => 'Only include modules specified in this list',
             'summary.alt.plurality.singular' => 'Add module to include list',
-            schema => ['array*', of=>'str*'],
+            schema => ['array*', of=>['str*']],
             element_completion => sub { _complete_module(@_, apply_filters=>0) },
             tags => ['category:filtering'],
         },
@@ -1166,7 +1319,7 @@ _
             'x.name.is_plural' => 1,
             summary => 'Exclude modules specified in this list',
             'summary.alt.plurality.singular' => 'Add module to exclude list',
-            schema => ['array*', of=>'str*'],
+            schema => ['array*', of=>['str*']],
             element_completion => sub { _complete_module(@_, apply_filters=>0) },
             tags => ['category:filtering'],
         },
@@ -1180,7 +1333,7 @@ _
             'x.name.is_plural' => 1,
             summary => 'Only include functions specified in this list',
             'summary.alt.plurality.singular' => 'Add function to include list',
-            schema => ['array*', of=>'str*'],
+            schema => ['array*', of=>['str*']],
             element_completion => sub { _complete_function(@_, apply_filters=>0) },
             tags => ['category:filtering'],
         },
@@ -1193,7 +1346,7 @@ _
             'x.name.is_plural' => 1,
             summary => 'Exclude functions specified in this list',
             'summary.alt.plurality.singular' => 'Add function to exclude list',
-            schema => ['array*', of=>'str*'],
+            schema => ['array*', of=>['str*']],
             element_completion => sub { _complete_function(@_, apply_filters=>0) },
             tags => ['category:filtering'],
         },
@@ -1207,7 +1360,7 @@ _
             'x.name.is_plural' => 1,
             summary => 'Only include participants whose seq/name matches this',
             'summary.alt.plurality.singular' => 'Add participant to include list',
-            schema => ['array*', of=>'str*'],
+            schema => ['array*', of=>['str*']],
             element_completion => sub { _complete_participant(@_, apply_filters=>0) },
             tags => ['category:filtering'],
         },
@@ -1225,7 +1378,7 @@ _
 You can specify `A & B` to include participants that have _both_ tags A and B.
 
 _
-            schema => ['array*', of=>'str*'],
+            schema => ['array*', of=>['str*']],
             element_completion => sub { _complete_participant_tags(@_, apply_filters=>0) },
             tags => ['category:filtering'],
         },
@@ -1233,7 +1386,7 @@ _
             'x.name.is_plural' => 1,
             summary => 'Exclude participants whose seq/name matches this',
             'summary.alt.plurality.singular' => 'Add participant to include list',
-            schema => ['array*', of=>'str*'],
+            schema => ['array*', of=>['str*']],
             element_completion => sub { _complete_participant(@_, apply_filters=>0) },
             tags => ['category:filtering'],
         },
@@ -1251,7 +1404,7 @@ _
 You can specify `A & B` to exclude participants that have _both_ tags A and B.
 
 _
-            schema => ['array*', of=>'str*'],
+            schema => ['array*', of=>['str*']],
             element_completion => sub { _complete_participant_tags(@_, apply_filters=>0) },
             tags => ['category:filtering'],
         },
@@ -1260,7 +1413,7 @@ _
             'x.name.is_plural' => 1,
             summary => 'Only include items whose seq/name matches this',
             'summary.alt.plurality.singular' => 'Add item to include list',
-            schema => ['array*', of=>'str*'],
+            schema => ['array*', of=>['str*']],
             element_completion => sub { _complete_item(@_, apply_filters=>0) },
             tags => ['category:filtering'],
         },
@@ -1273,7 +1426,7 @@ _
             'x.name.is_plural' => 1,
             summary => 'Exclude items whose seq/name matches this',
             'summary.alt.plurality.singular' => 'Add item to exclude list',
-            schema => ['array*', of=>'str*'],
+            schema => ['array*', of=>['str*']],
             element_completion => sub { _complete_item(@_, apply_filters=>0) },
             tags => ['category:filtering'],
         },
@@ -1287,26 +1440,26 @@ _
             'x.name.is_plural' => 1,
             summary => 'Only include datasets whose seq/name matches this',
             'summary.alt.plurality.singular' => 'Add dataset to include list',
-            schema => ['array*', of=>'str*'],
+            schema => ['array*', of=>['str*']],
             element_completion => sub { _complete_dataset(@_, apply_filters=>0) },
             tags => ['category:filtering'],
         },
         include_dataset_pattern => {
             summary => 'Only include datasets matching this regex pattern',
-            schema => 're*',
+            schema => ['re*'],
             tags => ['category:filtering'],
         },
         exclude_datasets => {
             'x.name.is_plural' => 1,
             summary => 'Exclude datasets whose seq/name matches this',
             'summary.alt.plurality.singular' => 'Add dataset to exclude list',
-            schema => ['array*', of=>'str*'],
+            schema => ['array*', of=>['str*']],
             element_completion => sub { _complete_dataset(@_, apply_filters=>0) },
             tags => ['category:filtering'],
         },
         exclude_dataset_pattern => {
             summary => 'Exclude datasets matching this regex pattern',
-            schema => 're*',
+            schema => ['re*'],
             tags => ['category:filtering'],
         },
         include_dataset_tags => {
@@ -1318,7 +1471,7 @@ _
 You can specify `A & B` to include datasets that have _both_ tags A and B.
 
 _
-            schema => ['array*', of=>'str*'],
+            schema => ['array*', of=>['str*']],
             element_completion => sub { _complete_dataset_tags(@_, apply_filters=>0) },
             tags => ['category:filtering'],
         },
@@ -1331,10 +1484,86 @@ _
 You can specify `A & B` to exclude datasets that have _both_ tags A and B.
 
 _
-            schema => ['array*', of=>'str*'],
+            schema => ['array*', of=>['str*']],
             element_completion => sub { _complete_dataset_tags(@_, apply_filters=>0) },
             tags => ['category:filtering'],
         },
+        include_dataset_tags => {
+            'x.name.is_plural' => 1,
+            summary => 'Only include datasets whose tag matches this',
+            'summary.alt.plurality.singular' => 'Add a tag to dataset include tag list',
+            description => <<'_',
+
+You can specify `A & B` to include datasets that have _both_ tags A and B.
+
+_
+            schema => ['array*', of=>['str*']],
+            element_completion => sub { _complete_dataset_tags(@_, apply_filters=>0) },
+            tags => ['category:filtering'],
+        },
+        multiperl => {
+            summary => 'Benchmark against multiple perls',
+            schema => ['bool'],
+            default => 0,
+            description => <<'_',
+
+Requires `App::perlbrew` to be installed. Will use installed perls from the
+perlbrew installation. Use `--include-perl` and `--exclude-perl` to include and
+exclude which perls you want.
+
+Each installed perl must have `Bencher::Backend` module installed (in addition
+to having all modules that you want to benchmark, obviously).
+
+Also note that due to the way this is currently implemented, benchmark code that
+contains closures (references to variables outside the code) won't work.
+
+_
+        },
+        include_perls => {
+            'x.name.is_plural' => 1,
+            'x.name.singular' => 'include_perl',
+            summary => 'Only include some perls',
+            'summary.alt.plurality.singular' => 'Add specified perl to include list',
+            schema => ['array*', of=>['str*']],
+            element_completion => sub { _complete_perl(@_) },
+            tags => ['category:filtering'],
+        },
+        exclude_perls => {
+            'x.name.is_plural' => 1,
+            'x.name.singular' => 'exclude_perl',
+            summary => 'Exclude some perls',
+            'summary.alt.plurality.singular' => 'Add specified perl to exclude list',
+            schema => ['array*', of=>['str*']],
+            element_completion => sub { _complete_perl(@_) },
+            tags => ['category:filtering'],
+        },
+
+        multimodver => {
+            summary => 'Benchmark multiple module versions',
+            schema => ['str*'],
+            description => <<'_',
+
+If set to a module name, will search for all (instead of the first occurrence)
+of the module in `@INC`. Then will generate items for each version.
+
+Currently only one module can be multi version.
+
+_
+            completion => sub { _complete_module(@_, apply_filters=>0) },
+        },
+        include_path => {
+            summary => 'Additional module search paths',
+            'summary.alt.plurality.singular' => 'Add path to module search path',
+            schema => ['array*', of=>['str*']],
+            description => <<'_',
+
+Only relevant when in multimodver mode.
+
+_
+            cmdline_aliases => {I=>{}},
+        },
+        # XXX include-mod-version
+        # XXX exclude-mod-version
 
         on_failure => {
             summary => "What to do when there is a failure",
@@ -1366,7 +1595,7 @@ _
         },
 
         sort => {
-            schema => ['array*', of=>'str*', min_len=>1],
+            schema => ['array*', of=>['str*'], min_len=>1],
             default => ['-time'],
         },
 
@@ -1382,12 +1611,12 @@ The default is to true (return extra metadata) unless when run as CLI and format
 is text (where the extra metadata is not shown).
 
 _
-            schema => 'bool',
+            schema => ['bool'],
         },
 
         note => {
             summary => 'Put additional note in the result',
-            schema => 'str*',
+            schema => ['str*'],
             tags => ['category:result'],
         },
     },
@@ -1511,6 +1740,21 @@ sub bencher {
         my @res;
         my $has_summary = 0;
         for my $p (@{ $parsed->{participants} }) {
+
+            my $cmdline;
+            if ($p->{cmdline_template}) {
+                $cmdline = "#TEMPLATE: ".
+                    (ref($p->{cmdline_template}) eq 'ARRAY' ? join(" ", @{$p->{cmdline_template}}) : $p->{cmdline_template});
+            } elsif ($p->{cmdline}) {
+                $cmdline =
+                    (ref($p->{cmdline}) eq 'ARRAY' ? join(" ", @{$p->{cmdline}}) : $p->{cmdline});
+            } elsif ($p->{perl_cmdline_template}) {
+                $cmdline = "#TEMPLATE: #perl ".
+                    (ref($p->{perl_cmdline_template}) eq 'ARRAY' ? join(" ", @{$p->{perl_cmdline_template}}) : $p->{perl_cmdline_template});
+            } elsif ($p->{cmdline}) {
+                $cmdline = "#perl ".
+                    (ref($p->{perl_cmdline}) eq 'ARRAY' ? join(" ", @{$p->{perl_cmdline}}) : $p->{perl_cmdline});
+            }
             my $rec = {
                 seq      => $p->{seq},
                 type     => $p->{type},
@@ -1518,7 +1762,7 @@ sub bencher {
                 name     => $p->{name} // $p->{_name},
                 function => $p->{function},
                 module   => $p->{module},
-                cmdline  => ref($p->{cmdline}) eq 'ARRAY' ? join(" ", @{$p->{cmdline}}) : $p->{cmdline},
+                cmdline  => $cmdline,
                 tags     => join(", ", @{$p->{tags} // []}),
             };
             if (defined $p->{summary}) {
@@ -1551,7 +1795,7 @@ sub bencher {
     my $items;
   GEN_ITEMS:
     {
-        if ($parsed->{item}) {
+        if ($parsed->{items}) {
             $items = $parsed->{items};
             last;
         }
@@ -1668,6 +1912,7 @@ sub bencher {
         my $on_result_failure = $args{on_result_failure} //
             $parsed->{on_result_failure} // $on_failure;
         {
+            last if $args{multiperl} || $args{multimodver};
             my $fitems = [];
             for my $it (@$items) {
                 $log->tracef("Testing code for item #%d (%s) ...",
@@ -1722,6 +1967,7 @@ sub bencher {
         }
 
         if ($action eq 'show-items-results') {
+            die "show-items-results currently not supported on multiperl or multimodver\n" if $args{multiperl} || $args{multimodver};
             if ($return_meta) {
                 $envres->[2] = [map {$_->{_result}} @$items];
             } elsif ($args{raw}) {
@@ -1772,42 +2018,103 @@ sub bencher {
         $envres->[3]{'func.precision'} = $precision if $return_meta;
 
         $log->tracef("Running benchmark (precision=%g) ...", $precision);
-        my $tres = Benchmark::Dumb::_timethese_guts(
-            $precision,
-            {
-                map { $_->{seq} => $_->{_code} } @$items
-            },
-            "silent",
-        );
-
-        if ($return_meta) {
-            $envres->[3]{'func.time_end'} = Time::HiRes::time();
-            $envres->[3]{'func.elapsed_time'} =
-                $envres->[3]{'func.time_end'} - $envres->[3]{'func.time_start'};
-            $envres->[3]{'func.sysload_after'} = [Sys::Load::getload()];
-        }
 
         my @columns = (qw/seq participant dataset/);
         my @rows;
-        for my $seq (sort {$a<=>$b} keys %$tres) {
-            my $it = _find_record_by_seq($items, $seq);
-            my $row = {
-                time    => $tres->{$seq}{result}{num},
-                rate    => 1 / $tres->{$seq}{result}{num},
-                samples => $tres->{$seq}{result}{_dbr_nsamples},
-                errors  => $tres->{$seq}{result}{errors}[0],
-                notes   => $it->{_code_error},
-            };
-
-            for my $k (sort keys %$it) {
-                next unless $k =~ /^(seq|participant|dataset|item_.+|arg_.+)$/;
-                push @columns, $k unless grep {$k eq $_} @columns;
-                $row->{$k} = $it->{$k};
+        if ($args{multiperl} || $args{multimodver}) {
+            require Data::Clone;
+            require File::Temp;
+            my %perl_exes;
+            my %perl_opts;
+            for my $it (@$items) {
+                $perl_exes{$it->{perl}} = $it->{_perl_exe};
+                $perl_opts{$it->{modver}} = $it->{_perl_opts} if defined $it->{modver};
+            }
+            if (!keys(%perl_opts)) {
+                $perl_opts{""} = [];
             }
 
-            push @rows, $row;
+            my $sc = Data::Clone::clone($parsed);
+            for (keys %$sc) { delete $sc->{$_} if /^(before|after)_/ } # remove all hooks
+
+            my $tempdir = File::Temp::tempdir(CLEANUP => $log->is_debug ? 0:1);
+
+            for my $perl (sort keys %perl_exes) {
+                for my $modver (sort keys %perl_opts) {
+                    my $scd_path = "$tempdir/scenario-$perl";
+                    $sc->{items} = [grep {
+                        $_->{perl} eq $perl &&
+                            (length($modver) ? $_->{modver} eq $modver : 1)
+                        } @$items];
+                    $log->debugf("Creating scenario dump file for %s (modver %s) at %s", $perl, $modver, $scd_path);
+                    open my($fh), ">", $scd_path or die "Can't open file $scd_path: $!";
+                    print $fh dmp($sc), ";\n";
+                    close $fh;
+                    my $res_path = "$tempdir/result-$perl";
+                    my $cmd = join(
+                        " ",
+                        $perl_exes{$perl},
+                        "-MBencher::Backend",
+                        "-MData::Dmp",
+                        @{ $perl_opts{$modver} },
+                        "-e'print dmp(Bencher::Backend::bencher(action=>q[bench], precision=>$precision, scenario_file=>q[$scd_path], return_meta=>0))' > '$res_path'",
+                    );
+                    $log->debugf("Running %s ...", $cmd);
+                    system $cmd;
+                    die "Failed running bencher for perl $perl (1)" if $?;
+                    my $res = do $res_path;
+                    die "Failed running bencher for perl $perl (2): can't parse result: $@" if $@;
+                    die "Failed running bencher for perl $perl (3): result not an enveloped result" if ref($res) ne 'ARRAY';
+                    die "Failed running bencher for perl $perl (4): $res->[0] - $res->[1]" if $res->[0] != 200;
+
+                    for my $row (@{ $res->[2] }) {
+                        $row->{perl} = $perl;
+                        push @columns, "perl" unless grep {$_ eq 'perl'} @columns;
+                        if (length $modver) {
+                            $row->{modver} = $modver;
+                            push @columns, "modver" unless grep {$_ eq 'modver'} @columns;
+                        }
+                        push @rows, $row;
+                    }
+                } # for modver
+            } # for perl
+        } else {
+            my $tres = Benchmark::Dumb::_timethese_guts(
+                $precision,
+                {
+                    map { $_->{seq} => $_->{_code} } @$items
+                },
+                "silent",
+            );
+
+            if ($return_meta) {
+                $envres->[3]{'func.time_end'} = Time::HiRes::time();
+                $envres->[3]{'func.elapsed_time'} =
+                    $envres->[3]{'func.time_end'} - $envres->[3]{'func.time_start'};
+                $envres->[3]{'func.sysload_after'} = [Sys::Load::getload()];
+            }
+
+            for my $seq (sort {$a<=>$b} keys %$tres) {
+                my $it = _find_record_by_seq($items, $seq);
+                my $row = {
+                    time    => $tres->{$seq}{result}{num},
+                    rate    => 1 / $tres->{$seq}{result}{num},
+                    samples => $tres->{$seq}{result}{_dbr_nsamples},
+                    errors  => $tres->{$seq}{result}{errors}[0],
+                    notes   => $it->{_code_error},
+                };
+
+                for my $k (sort keys %$it) {
+                    next unless $k =~ /^(seq|participant|dataset|item_.+|arg_.+)$/;
+                    push @columns, $k unless grep {$k eq $_} @columns;
+                    $row->{$k} = $it->{$k};
+                }
+                push @rows, $row;
+            }
         }
+
         push @columns, qw/seq rate time errors samples notes/;
+
         $envres->[2] = \@rows;
         $envres->[3]{'table.fields'} = \@columns;
 
@@ -1897,7 +2204,7 @@ Bencher::Backend - Backend for Bencher
 
 =head1 VERSION
 
-This document describes version 0.38 of Bencher::Backend (from Perl distribution Bencher-Backend), released on 2016-03-15.
+This document describes version 0.39 of Bencher::Backend (from Perl distribution Bencher-Backend), released on 2016-03-15.
 
 =head1 FUNCTIONS
 
@@ -1972,6 +2279,10 @@ You can specify C<A & B> to exclude participants that have I<both> tags A and B.
 
 Exclude participants whose seq/name matches this.
 
+=item * B<exclude_perls> => I<array[str]>
+
+Exclude some perls.
+
 =item * B<include_dataset_pattern> => I<re>
 
 Only include datasets matching this regex pattern.
@@ -2024,10 +2335,43 @@ You can specify C<A & B> to include participants that have I<both> tags A and B.
 
 Only include participants whose seq/name matches this.
 
+=item * B<include_path> => I<array[str]>
+
+Additional module search paths.
+
+Only relevant when in multimodver mode.
+
+=item * B<include_perls> => I<array[str]>
+
+Only include some perls.
+
 =item * B<module_startup> => I<bool>
 
 Benchmark module startup overhead instead of normal benchmark.
 
+=item * B<multimodver> => I<str>
+
+Benchmark multiple module versions.
+
+If set to a module name, will search for all (instead of the first occurrence)
+of the module in C<@INC>. Then will generate items for each version.
+
+Currently only one module can be multi version.
+
+=item * B<multiperl> => I<bool> (default: 0)
+
+Benchmark against multiple perls.
+
+Requires C<App::perlbrew> to be installed. Will use installed perls from the
+perlbrew installation. Use C<--include-perl> and C<--exclude-perl> to include and
+exclude which perls you want.
+
+Each installed perl must have C<Bencher::Backend> module installed (in addition
+to having all modules that you want to benchmark, obviously).
+
+Also note that due to the way this is currently implemented, benchmark code that
+contains closures (references to variables outside the code) won't work.
+
 =item * B<note> => I<str>
 
 Put additional note in the result.
