@@ -1,7 +1,7 @@
 package Bencher::Formatter::DeleteConstantFields;
 
 our $DATE = '2016-03-15'; # DATE
-our $VERSION = '0.38'; # VERSION
+our $VERSION = '0.39'; # VERSION
 
 use 5.010001;
 use strict;
@@ -23,7 +23,7 @@ sub munge_result {
         last unless $td->row_count >= 2;
         my @const_cols = $td->const_col_names;
         for my $k (@const_cols) {
-            next unless $k =~ /^(item_.+|arg_.+|participant|dataset)$/;
+            next unless $k =~ /^(item_.+|arg_.+|perl|modver|participant|dataset)$/;
             for my $row (@{ $envres->[2] }) {
                 delete $row->{$k};
             }
@@ -46,7 +46,7 @@ Bencher::Formatter::DeleteConstantFields - Delete constant fields to reduce clut
 
 =head1 VERSION
 
-This document describes version 0.38 of Bencher::Formatter::DeleteConstantFields (from Perl distribution Bencher-Backend), released on 2016-03-15.
+This document describes version 0.39 of Bencher::Formatter::DeleteConstantFields (from Perl distribution Bencher-Backend), released on 2016-03-15.
 
 =head1 DESCRIPTION
 
